import struct
import numpy as np
from mdio.base import Frame

def read_record(f, expectedbytes=None, id='data'):
    nbytes = struct.unpack('i', f.read(4))[0]
    buf = f.read(nbytes)
    check = struct.unpack('i', f.read(4))[0]
    if check != nbytes:
        raise ValueError('Error - malformed record in DCD file.')
    if expectedbytes is not None:
        if nbytes != expectedbytes:
            raise IOError('Error reading {} from DCD file.'.format(id))
    return buf

def write_record(f, buf):
    nbytes = len(buf)
    f.write(struct.pack('i', nbytes))
    f.write(struct.pack('{}s'.format(nbytes), buf))
    f.write(struct.pack('i', nbytes))

class DCDHeader(object):
    """
    The header block for a DCD file.
    """
    def __init__(self, istart=1, nsavec=0, deltat=1.0, version=24, extra=0):
        self.istart = istart
        self.nsavec = nsavec
        self.deltat = deltat
        self.version = version
        self.nsets = 0
        self.extra = extra


def read_dcd_header(f):
    f.seek(0)
    buf = read_record(f, 84, 'header')
    magic = struct.unpack('4s', buf[:4])[0]
    if magic.decode('utf-8') != 'CORD':
        raise IOError('Error - this does not seem to be a DCD format file ({}).'.format(magic))
    header = DCDHeader()
    header.version = struct.unpack('i', buf[80:])[0]
    header.extra = struct.unpack('i', buf[44:48])[0]
    header.nsets, header.istart, header.nsavec = struct.unpack('3i', buf[8:20])
    header.deltat = struct.unpack('f', buf[40:44])[0]
    return header

def write_dcd_header(f, header):
    f.seek(0)
    buf = bytearray(84)
    buf[:4] = struct.pack('4s', b'CORD')
    buf[8:20] = struct.pack('3i', header.nsets, header.istart, header.nsavec)
    buf[40:48] = struct.pack('fi', header.deltat, header.extra)
    buf[80:] = struct.pack('i', header.version)
    write_record(f, buf)
    
def update_dcd_header(f, nsavec, deltat=None):
    here = f.tell()
    f.seek(20)
    f.write(struct.pack('i', nsavec))
    if deltat is not None:
        f.seek(44)
        f.write(struct.pack('f', deltat))
    f.seek(here)

def read_dcd_titles(f):
    buf = read_record(f)
    ntitle = struct.unpack('i', buf[:4])[0]
    if (len(buf) - 4) // 80 != ntitle:
        raise IOError('Error - cannot read title info from DCD file.')
    titles = []
    for i in range(ntitle):
        offset = 4 + (80 * i)
        titles.append(struct.unpack('80s', buf[offset:offset+80])[0])
    return titles

def write_dcd_titles(f, titles):
    ntitles = len(titles)
    buf = bytearray(4 + 80 * ntitles)
    buf[:4] = struct.pack('i', ntitles)
    for i in range(ntitles):
        offset = 4 + (80 * i)
        title = titles[i]
        lt = len(title)
        lt = min(lt, 80)
        if isinstance(title, str):
            title = bytes(title[:lt], 'utf-8')
        
        buf[offset:offset+lt] = struct.pack('{}s'.format(lt), title)
    write_record(f, buf)

def read_dcd_natoms(f):
    buf = read_record(f, 4, 'natoms')
    natoms = struct.unpack('i', buf)[0]
    return natoms

def write_dcd_natoms(f, natoms):
    buf = struct.pack('i', natoms)
    write_record(f, buf)
    
def read_dcd_unit_cell(f):
    buf = read_record(f, 48, 'box data')
    unit_cell_info = struct.unpack('6d', buf)
    return unit_cell_info

def write_dcd_unit_cell(f, unit_cell_data):
    if len(unit_cell_data) != 6:
        raise IOError('Error - must be 6 values in unit cell data list.')
    buf = struct.pack('6d', *unit_cell_data)
    write_record(f, buf)

def read_dcd_coordinates(f, natoms):
    buf = read_record(f, natoms * 4, 'x-coordinates')
    x = struct.unpack('{}f'.format(natoms), buf)
    buf = read_record(f, natoms * 4, 'y-coordinates')
    y = struct.unpack('{}f'.format(natoms), buf)
    buf = read_record(f, natoms * 4, 'z-coordinates')
    z = struct.unpack('{}f'.format(natoms), buf)
    return [list(c) for c in zip(x,y,z)]

def write_dcd_coordinates(f, crds):
    x = [c[0] for c in crds]
    natoms = len(x)
    xbuf = struct.pack('{}f'.format(natoms), *x)
    write_record(f, xbuf)
    y = [c[1] for c in crds]
    write_record(f, struct.pack('{}f'.format(natoms), *y))
    z = [c[2] for c in crds]
    write_record(f, struct.pack('{}f'.format(natoms), *z))

class DCDFileReader(object):
    def __init__(self, f):
        self.f  = f
        self.header = read_dcd_header(f)
        self.titles = read_dcd_titles(f)
        self.natoms = read_dcd_natoms(f)
        self.hasbox = self.header.extra == 1
        self.framecount = 0
        self.nframes = self.header.nsavec
        
    def __del__(self):
        try:
            self.f.close()
        except:
            pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()
    
    def close(self):
        self.f.close()
        
    def read_frame(self):
        if self.framecount >= self.nframes:
            return None
        if self.hasbox:
            box = read_dcd_unit_cell(self.f)
        else:
            box = None
        crds = read_dcd_coordinates(self.f, self.natoms)
        self.framecount += 1
        timestep = self.header.istart + self.framecount - 1
        time = timestep * self.header.deltat
        frame = Frame(crds, box, time=time, timestep=timestep)
        return frame
        
class DCDFileWriter(object):
    def __init__(self, f, titles=['Created by DCDFileWriter']):
        self.f = f
        self.header = None
        self.titles = titles
        self.natoms = None
        self.hasbox = None
        self.framecount = 0
        self.deltat = None
        
    def __del__(self):
        try:
            self.close()
        except:
            pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()
        
    def close(self):
        if self.header is not None:
            if self.deltat < 0:
                self.deltat = 1.0
            update_dcd_header(self.f, self.framecount, self.deltat)
        self.f.close()
        
        
    def write_frame(self, frame):
        if self.natoms is None:
            self.natoms = frame.natoms
            self.hasbox = frame.box is not None
        elif self.natoms != frame.natoms:
            raise IOError('Error - expected {} atoms in frame, found {}.'.format(self.natoms, frame.natoms))
        
        if self.hasbox and (frame.box is None):
            raise IOError('Error - frame is missing box data.')
        if not self.hasbox and (frame.box is not None):
            raise IOError('Error - frame has unexpected box data.')
            
        if self.header is None:
            self.header = DCDHeader()
            if self.hasbox:
                self.header.extra = 1
            else:
                self.header.extra = 0
            write_dcd_header(self.f, self.header)
            write_dcd_titles(self.f, self.titles)
            write_dcd_natoms(self.f, self.natoms)
        if self.deltat is None:
            self.deltat = -frame.time
        elif self.deltat < 0:
            self.deltat += frame.time
        if self.hasbox:
            box = frame.box
            tbox = [box[0,0]]
            tbox.append(box[1,0])
            tbox.append(box[1,1])
            tbox.append(box[2,0])
            tbox.append(box[2,1])
            tbox.append(box[2,2])
            write_dcd_unit_cell(self.f, tbox)
        write_dcd_coordinates(self.f, frame.crds)
        self.framecount += 1

def dcd_open(filename, mode='r'):
    if not mode in ['r', 'w']:
        raise ValueError('Error - mode must be "r" or "w".')
    if mode == 'r':
        f = open(filename, 'rb')
        filehandler = DCDFileReader(f)
    elif mode == 'w':
        f = open(filename, 'wb')
        filehandler = DCDFileWriter(f)
    return filehandler



import numpy as np
class Frame(object):
    """
    A frame of trajectory data.
    
    """
    def __init__(self, crds, box=None, time=0.0, precision=1000, timestep = 1.0):
        crds = np.array(crds, dtype=np.float32)
        if len(crds.shape) != 2:
            raise TypeError('Error - crds must be a [N,3] array.')
        if crds.shape[1] != 3:
            raise TypeError('Error - crds must be a [N,3] array.')
        self.natoms = crds.shape[0]
        self.crds = crds
        if box is not None:
            box = np.array(box, dtype=np.float32)
            if len(box.shape) == 1 and len(box) == 6:
                tbox = np.zeros((3,3), dtype=np.float32)
                tbox[0, 0] = box[0]
                tbox[1, 0] = box[1]
                tbox[1, 1] = box[2]
                tbox[2, 0] = box[3]
                tbox[2, 1] = box[4]
                tbox[2, 2] = box[5]
                box = tbox
            elif box.shape != (3,3):
                raise ValueError('Error - unrecognised box data {}'.format(box))
        self.box = box
        self.time = float(time)
        self.precision = int(precision)
        self.timestep = timestep



from mdio.ncio import nc_open
import numpy as np

ncfile = 'examples/test.nc'
testcrds = np.array([28.37, 43.48, 25.27])
testbox = np.array([[59.414,  0.,     0.   ],
     [ 0.,    59.414,  0.,   ],
     [ 0.,     0.,    59.414]], dtype=np.float32)
testi = 220

f = nc_open(ncfile, 'r')
frame = f.read_frame()
assert np.allclose(frame.crds[0], testcrds) is True
assert np.allclose(frame.box, testbox) is True
i = 1
while frame is not None:
    frame = f.read_frame()
    i += 1
assert i == testi

f.close()
fin = nc_open(ncfile, 'r')
fout = nc_open('tmp.nc', 'w')
frame = fin.read_frame()
while frame is not None:
    fout.write_frame(frame)
    frame = fin.read_frame()
fin.close()
fout.close()

finA = nc_open(ncfile, 'r')
finB = nc_open('tmp.nc', 'r')
frameA = finA.read_frame()
frameB = finB.read_frame()
while frameA is not None:
    assert np.allclose(frameA.crds, frameB.crds) is True
    frameA = finA.read_frame()
    frameB = finB.read_frame()
finA.close()
finB.close()

ncfile = 'examples/test_nobox.nc'

f = nc_open(ncfile, 'r')
frame = f.read_frame()
assert np.allclose(frame.crds[0], testcrds) is True
i = 1
while frame is not None:
    frame = f.read_frame()
    i += 1
assert i == testi

f.close()
fin = nc_open(ncfile, 'r')
fout = nc_open('tmp.nc', 'w')
frame = fin.read_frame()
while frame is not None:
    fout.write_frame(frame)
    frame = fin.read_frame()
fin.close()
fout.close()

finA = nc_open(ncfile, 'r')
finB = nc_open('tmp.nc', 'r')
frameA = finA.read_frame()
frameB = finB.read_frame()
while frameA is not None:
    assert np.allclose(frameA.crds, frameB.crds) is True
    frameA = finA.read_frame()
    frameB = finB.read_frame()
finA.close()
finB.close()




from mdio.xtcio import xtc_open
import numpy as np

testbox = np.array(
    [[5.94140005, 0.,         0.        ],
     [0.,         5.94140005, 0.        ],
     [0.,         0.,         5.94140005]]
    )
xtcfile = 'examples/test.xtc'
testcrds = np.array([2.837, 4.348, 2.527]) 
testi = 220

f = xtc_open(xtcfile, 'r')
frame = f.read_frame()
assert np.allclose(frame.crds[0], testcrds) is True
assert np.allclose(frame.box, testbox) is True
i = 1
while frame is not None:
    frame = f.read_frame()
    i += 1
assert i == testi
f.close()

fin = xtc_open(xtcfile, 'r')
fout = xtc_open('tmp.xtc', 'w')
frame = fin.read_frame()
while frame is not None:
    fout.write_frame(frame)
    frame = fin.read_frame()
fin.close()
fout.close()

finA = xtc_open(xtcfile, 'r')
finB = xtc_open('tmp.xtc', 'r')
frameA = finA.read_frame()
frameB = finB.read_frame()
i = 0
while frameA is not None:
    if not np.allclose(frameA.crds, frameB.crds):
        print(i, frameA.crds[-1],frameB.crds[-1])
    #assert np.allclose(frameA.crds, frameB.crds) is True
    frameA = finA.read_frame()
    frameB = finB.read_frame()
    i += 1
finA.close()
finB.close()

xtcfile = 'examples/test_nobox.xtc'
testi = 220

f = xtc_open(xtcfile, 'r')
frame = f.read_frame()
assert np.allclose(frame.crds[0], testcrds) is True
i = 1
while frame is not None:
    frame = f.read_frame()
    i += 1
assert i == testi
f.close()

fin = xtc_open(xtcfile, 'r')
fout = xtc_open('tmp.xtc', 'w')
frame = fin.read_frame()
while frame is not None:
    fout.write_frame(frame)
    frame = fin.read_frame()
fin.close()
fout.close()

finA = xtc_open(xtcfile, 'r')
finB = xtc_open('tmp.xtc', 'r')
frameA = finA.read_frame()
frameB = finB.read_frame()
i = 0
while frameA is not None:
    if not np.allclose(frameA.crds, frameB.crds):
        print(i, frameA.crds[-1],frameB.crds[-1])
    assert np.allclose(frameA.crds, frameB.crds) is True
    frameA = finA.read_frame()
    frameB = finB.read_frame()
    i += 1
finA.close()
finB.close()




from mdio.xtcio import xtc_open
import time

xtcfile = 'examples/test.xtc'

frames = []
f = xtc_open(xtcfile, 'r')
time0 = time.time()
frame = f.read_frame()
while frame is not None:
    frame = f.read_frame()
    frames.append(frame)
f.close()
time1 = time.time()

fout = xtc_open('tmp.xtc', 'w')
time2 = time.time()
for frame in frames[:-1]:
    fout.write_frame(frame)
fout.close()
time3 = time.time()

print('time to read:', time1-time0)
print('time to write:', time3-time2)



# Reading and writing Gromacs xtc files
#
# Sections of the code here has been rewritten or adapted from the c code in 
# the Gromacs source file libxdrf.cpp.
#
# Part of the header to that file is reproduced here:
#/*
# * This file is part of the GROMACS molecular simulation package.
# *
# * Copyright (c) 1991-2000, University of Groningen, The Netherlands.
# * Copyright (c) 2001-2004, The GROMACS development team.
# * Copyright (c) 2013,2014,2015,2016,2017,2018, by the GROMACS development team, led by
# * Mark Abraham, David van der Spoel, Berk Hess, and Erik Lindahl,
# * and including many others, as listed in the AUTHORS file in the
# * top-level source directory and at http://www.gromacs.org.
# *
#
import xdrlib
import numpy as np
from mdio.xtcutils import BitByteBuffer
from mdio.base import Frame

def sizeofint(imax):
    """
    Number of bits required to store an integer of max size imax.
    """
    num_of_bits = int(imax).bit_length()
    return num_of_bits

def sizeofints(sizes):
    """
    Number of bits required to pack a list of integers with max values sizes.
    
    """
    p = 1
    for s in sizes:
        p = p * s
    i = int(p).bit_length()
    return i

class MagicStuff(object):
    
    def __init__(self, smallidx=None, mindif=None):
        
        self.magicints = [
        0, 0, 0, 0, 0, 0, 0, 0, 0,
        8, 10, 12, 16, 20, 25, 32, 40, 50, 64,
        80, 101, 128, 161, 203, 256, 322, 406, 512, 645,
        812, 1024, 1290, 1625, 2048, 2580, 3250, 4096, 5060, 6501,
        8192, 10321, 13003, 16384, 20642, 26007, 32768, 41285, 52015, 65536,
        82570, 104031, 131072, 165140, 208063, 262144, 330280, 416127, 524287, 660561,
        832255, 1048576, 1321122, 1664510, 2097152, 2642245, 3329021, 4194304, 5284491, 6658042,
        8388607, 10568983, 13316085, 16777216
        ]

        self.FIRSTIDX = 9
        self.LASTIDX = len(self.magicints)
        
        if smallidx is None and mindif is None:
            raise ValueError('Must provide either smallidx or mindif')
    
        if smallidx is None:
            smallidx = self.FIRSTIDX
            while (smallidx < self.LASTIDX) and (self.magicints[smallidx] < mindif):
                smallidx += 1
    
        self.smallidx = smallidx
        self.maxidx = min(self.LASTIDX, smallidx + 8)
        self.minidx = self.maxidx - 8
        self.larger = self.magicints[self.maxidx] // 2
        self.smaller = self.magicints[max(self.FIRSTIDX, smallidx-1)] // 2
        self.smallnum = self.magicints[smallidx] // 2
        self.sizesmall = np.array([self.magicints[smallidx]] * 3, dtype=np.int)
            
    def update(self, is_smaller):
    
        self.smallidx += is_smaller
        if is_smaller < 0:
            self.smallnum = self.smaller
            if self.smallidx > self.FIRSTIDX:
                self.smaller = self.magicints[self.smallidx - 1] // 2
            else:
                self.smaller = 0
        elif is_smaller > 0:
            self.smaller  = self.smallnum
            self.smallnum = self.magicints[self.smallidx] // 2
        self.sizesmall = np.array([self.magicints[self.smallidx]] * 3, dtype=np.int)

def read_frame(up, natoms=None):
    """
    Read a frame of trajectory data out of an xdrlib unpacker buffer.
    
    """
    try:
        magic = up.unpack_int()
    except:
        return None
    if magic != 1995:
        raise ValueError('Error reading frame: magic is {}'.format(magic))
        
    maxatoms = up.unpack_int()
    timestep = up.unpack_int()
    time = up.unpack_float()
    box = np.array(up.unpack_farray(9, up.unpack_float)).reshape((3,3))    
    natoms_check = up.unpack_int()
    
    if natoms_check != maxatoms:
        raise ValueError('Error - frame has {} coordinates instead of {}'.format(natoms_check, natoms))
        
    if natoms is None:
        natoms = maxatoms
    elif natoms > maxatoms:
        raise ValueError('Error - frame only contains {} atoms'.format(maxatoms))
    if maxatoms <= 9:
        crds = np.array(up.unpack_farray(maxatoms*3, up.unpack_float)).reshape((maxatoms, 3))
        frame = Frame(crds[:natoms], box=box, time=time, timestep=timestep)
        return frame
                               
    precision = up.unpack_float()
    if precision == 0:
        return None
    inv_precision = 1.0 / precision
    
    minint = np.array(up.unpack_farray(3, up.unpack_int), dtype=np.int)
    maxint = np.array(up.unpack_farray(3, up.unpack_int), dtype=np.int)
    smallidx = up.unpack_int()
    sizes = maxint - minint + 1
    bitsizeint = np.zeros(3, dtype=np.int)
    
    if any(sizes > 0xffffff):
        bitsizeint[0] = sizeofint(sizes[0])
        bitsizeint[1] = sizeofint(sizes[1])
        bitsizeint[2] = sizeofint(sizes[2])
        bitsize = 0
    else:
        bitsize = sizeofints(sizes)
        
    ms = MagicStuff(smallidx=smallidx)
    crds = np.zeros((maxatoms, 3))
    
    cnt = up.unpack_int()

    ba = bytearray(up.unpack_fopaque(cnt))
    bs = BitByteBuffer(ba)
    i = 0
    run = 0
    while i < natoms:
        if bitsize == 0:
            thiscoord = np.zeros(3, dtype=np.int)
            for j in range(3):
                thiscoord[j]= bs.receiveint(bitsizeint[j])
        else:
            nums = bs.receiveints(bitsize, sizes)
            thiscoord = np.array(nums, dtype=np.int)
        thiscoord += minint
        prevcoord = thiscoord

        flag = bs.receiveint(1)
        is_smaller = 0
        if flag == 1:
            run = bs.receiveint(5)
            is_smaller = run % 3
            run -= is_smaller
            is_smaller -= 1
        
        if run > 0:
            for k in range(0, run, 3):
                nums = bs.receiveints(ms.smallidx, ms.sizesmall)
                thiscoord = np.array(nums, dtype=np.int)
                thiscoord += prevcoord - ms.smallnum
                if (k == 0):
                    tmp = thiscoord
                    thiscoord = prevcoord
                    prevcoord = tmp
                    crds[i] = prevcoord
                    i += 1
                else:
                    prevcoord = thiscoord
                crds[i] = thiscoord
                i += 1
        else:
            crds[i] = thiscoord
            i += 1
        
        ms.update(is_smaller)
        
    frame = Frame(crds[:natoms] * inv_precision, box=box, time=time, timestep=timestep, precision=precision)
    return frame


def write_frame(p, frame):
    """
    Write a trajectory frame to the xdrlib packer buffer.
    
    """
    INT_MAX = 2147483647
    INT_MIN = -2147483648
    MAXABS = INT_MAX - 2
    MAGIC = 1995
    
    p.pack_int(MAGIC)
    p.pack_int(frame.natoms)
    p.pack_int(frame.timestep)
    p.pack_float(frame.time)
    if frame.box is None:
        p.pack_farray(9, np.zeros((9)), p.pack_float)
    else:
        p.pack_farray(9, frame.box.flatten(), p.pack_float)  
    
    crds = frame.crds
    size = len(crds)
    p.pack_int(size)
    size3 = size * 3
    if size <= 3:
        p.pack_farray(size3, crds.flatten(), p.pack_float)
        return
    p.pack_float(frame.precision)
    precision = frame.precision
    
    ba = bytearray(size3*8)
    buf = BitByteBuffer(ba)
    prevrun = -1
    
    if abs(crds * precision + 0.5).max() > MAXABS:
        raise ValueError('Error - coordinate values too great for precision={}'.format(precision))
    crdrange = crds.max(axis=0) - crds.min(axis=0)
    if crdrange.max() > MAXABS:
        raise ValueError('Error - coordinate range to great for precision={}'.format(precision))
    
    icrds = np.rint((crds * precision)).astype(np.int)
    minint = icrds.min(axis=0)
    maxint = icrds.max(axis=0)
    sizeint = maxint - minint + 1
    
    for i in range(3):
        p.pack_int(minint[i])

    for i in range(3):
        p.pack_int(maxint[i])

    
    bitsizeint = np.zeros(3, dtype=np.int)
    if any(sizeint > 0xffffff):
        bitsizeint[0] = sizeofint(sizeint[0])
        bitsizeint[1] = sizeofint(sizeint[1])
        bitsizeint[2] = sizeofint(sizeint[2])
        bitsize = 0
    else:
        bitsize = sizeofints(sizeint)


    mindif = np.abs(icrds[1:] - icrds[:-1]).sum(axis=1).min()
    ms = MagicStuff(mindif=mindif)
    p.pack_int(ms.smallidx)
        
    i = 0
    prevcoord = np.zeros(3, dtype=np.int)
    tmpcoord = np.zeros(30, dtype=np.intc)
    tmp = np.zeros(3, dtype=np.int)
    while i < size:
        is_small  = False

        if (ms.smallidx < ms.maxidx) and (i >= 1) and all(np.abs(icrds[i] - prevcoord) < ms.larger):
            is_smaller = 1
        elif ms.smallidx > ms.minidx:
            is_smaller = -1
        else:
            is_smaller = 0
            
        if i + 1 < size:
            if all(np.abs(icrds[i] - icrds[i+1]) < ms.smallnum):
                tmp[:] = icrds[i]
                icrds[i] = icrds[i+1]
                icrds[i+1] = tmp[:]
                is_small = True

        tmpcoord[:3] = icrds[i] - minint
        if bitsize == 0:   
            buf.sendint(bitiszeint[0], tmpcoord[0])
            buf.sendint(bitiszeint[1], tmpcoord[1])
            buf.sendint(bitiszeint[2], tmpcoord[2])
        else:
            buf.sendints(bitsize, sizeint, tmpcoord[:3])
        prevcoord[:] = icrds[i]
        i += 1
        
        run = 0
        if not is_small and is_smaller == -1:
            is_smaller = 0
    
        while is_small and run < 8*3:
            dif = icrds[i] - prevcoord
            if is_smaller == -1 and (dif*dif).sum() >= (ms.smaller * ms.smaller):
                is_smaller = 0

            tmpcoord[run:run+3] = dif + ms.smallnum;
            run += 3
            prevcoord[:] = icrds[i]
            i += 1
            if i >= size:
                is_small = False
            else:
                dif = icrds[i] - prevcoord
                is_small = all(np.abs(dif) < ms.smallnum) 
            
        if run != prevrun or is_smaller != 0:
            prevrun = run           
            buf.sendint(1, 1)
            buf.sendint(5, run+is_smaller+1)
        else:
            buf.sendint(1, 0)
        for k in range(0, run, 3):
            buf.sendints(ms.smallidx, ms.sizesmall, tmpcoord[k:k+3])

        ms.update(is_smaller)
                            
    bout = buf.tobytes()
    p.pack_int(len(bout))
    p.pack_fopaque(len(bout), bout)

def xtc_open(filename, mode='r'):
    if not mode in ['r', 'w']:
        raise ValueError('Error: mode must be "r" or "w"')
    if mode == 'r':
        filehandler = XtcfileReader(filename)
    else:
        filehandler = XtcfileWriter(filename)
    return filehandler

class XtcfileReader(object):
    """
    Reader for xtc format files.

    """
    def __init__(self, filename):
        self.filename = filename
        with open(self.filename, 'rb') as f:
            self.up =  xdrlib.Unpacker(f.read())

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def read_frame(self, natoms=None):
        """
        Returns the next frame from the xtc file.

        """
        return read_frame(self.up, natoms=natoms)

    def close(self):
        """
        Close the file (a no-op in read mode)

        """
        pass


class XtcfileWriter(object):
    """
    Writer for xtc format files.

    """
    def __init__(self, filename):
        self.filename = filename
        self.f = open(self.filename, 'wb')
        self.p = xdrlib.Packer()

    def __del__(self):
        try:
            self.f.close()
        except:
            pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        """
        Close the file 

        """
        self.f.close()

    def write_frame(self, frame):
        """
        Write a frame of data to the file.

        """
        self.p.reset()
        if frame.crds is None:
            raise ValueError('Error - no coordinates in frame')
        frame.natoms = len(frame.crds)
        if frame.precision is None:
            frame.precision = 1000
        if frame.box is None:
            frame.box = np.zeros((3,3))
        if frame.time is None:
            frame.time = self.time
            self.time += 1.0
        else:
            self.time = frame.time + 1.0
        if frame.timestep is None:
            frame.timestep = self.timestep
            self.timestep += 1
        else:
            self.timestep = frame.timestep + 1
        write_frame(self.p, frame)
        self.f.write(self.p.get_buffer())




from setuptools import setup, find_packages
from setuptools.extension import Extension
from Cython.Build import cythonize
import re
import numpy

VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
VERSIONFILE = "mdio/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RunTimeError("Unable to find version string in {}.".format(VERSIONFILE))

with open("README.md", "r") as f:
    long_description = f.read()

extensions = [Extension(
        name="mdio.xtcutils",
        sources=["mdio/src/xtcutils.pyx"],
        include_dirs=[numpy.get_include()],
        )
    ]
setup(
    name = 'mdio',
    version = verstr,
    author = 'Charlie Laughton',
    author_email = 'charles.laughton@nottingham.ac.uk',
    description = 'Basic I/O for MD trajectory files',
    long_description = long_description,
    long_description_content_type='text/markdown',
    url = 'https://bitbucket.org/claughton/mdio',
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Development Status :: 3 - Alpha",
    ],
    packages = find_packages(),
    ext_modules = cythonize(extensions),
    install_requires = [
        'numpy',
        'cython',
    ],
)



from netCDF4 import Dataset
import numpy as np
from mdio.base import Frame

class NCFileReader(object):
    def __init__(self, filename):
        self.filename = filename
        self.root = Dataset(filename, 'r')
        if self.root.Conventions != "AMBER":
            raise TypeError("Error - this does not appear" 
                             " to be an Amber netcdf file.")
        self.coordinates = self.root['/coordinates']
        self.time = self.root['/time']
        self.index = -1
        self.nframes = len(self.coordinates)
        self.periodic = 'cell_spatial' in self.root.dimensions
        if self.periodic:
            self.cell_lengths = self.root['/cell_lengths']
            self.cell_angles = self.root['/cell_angles']
            
    def read_frame(self):
        self.index += 1
        if self.index >= self.nframes:
            return None
        if self.periodic:
            box = la2v(self.cell_lengths[self.index], 
                       self.cell_angles[self.index])
        else:
            box = None
        frame = Frame(self.coordinates[self.index], 
                      box=box, 
                      time=self.time[self.index])
        return frame
    
    def close(self):
        self.root.close()
        
    def __enter__(self):
        return self
    
    def __exit__(self, *args):
        self.close()
        
class NCFileWriter(object):
    def __init__(self, filename):
        self.filename = filename
        self.index = -1
        self.root = Dataset(filename, 'w', format="NETCDF3_64BIT_OFFSET")
        self.root.set_fill_off()
        
        self.root.Conventions = "AMBER"
        self.root.ConventionVersion = "1.0"
        self.root.program = "mdio"
        self.root.programVersion = "0.0.2"
        self.root.title = "CREATED by mdio.ncio"
        
        self.root.createDimension("frame", None)
        self.root.createDimension("spatial", 3)
        
        self.spatial = self.root.createVariable("spatial", "S1", ("spatial",))
        self.spatial[0] = "x"
        self.spatial[1] = "y"
        self.spatial[2] = "z"

        self.time = self.root.createVariable("time", "f4", ("frame",))
        self.time.units = "picoseconds"
        
        self.periodic = False

    def set_coordinates(self, natoms):
        self.root.createDimension("atom", natoms)
        self.coordinates = self.root.createVariable("coordinates", "f4", 
                                                    ("frame", "atom", "spatial"))
        self.coordinates.units = "angstrom"
        
    def set_periodic(self):
        if 'cell_spatial' in self.root.dimensions:
            return
        self.root.createDimension("cell_spatial", 3)
        self.root.createDimension("cell_angular", 3)
        self.root.createDimension("label", 5)
        
        self.cell_angular = self.root.createVariable("cell_angular", "S1", 
                                                     ("cell_spatial", "label"))
        self.cell_spatial = self.root.createVariable("cell_spatial", "S1", 
                                                     ("cell_spatial",))
                
        self.cell_angular[0] = "alpha"
        self.cell_angular[1] = "beta"
        self.cell_angular[2] = "gamma"

        self.cell_spatial[0] = "a"
        self.cell_spatial[1] = "b"
        self.cell_spatial[2] = "c"

        self.cell_lengths = self.root.createVariable("cell_lengths", "f4", 
                                                     ("frame", "cell_spatial"))
        self.cell_lengths.units = "angstrom"

        self.cell_angles = self.root.createVariable("cell_angles", "f4", 
                                                    ("frame", "cell_angular"))
        self.cell_angles.units = "degree"
        
        self.periodic = True
        
    def write_frame(self, frame):
        self.index += 1
        self.time[self.index] = frame.time
        if self.index == 0:
            self.set_coordinates(frame.natoms)
            if frame.box is not None:
                self.set_periodic()

        self.coordinates[self.index] = frame.crds
        if frame.box is None and self.periodic:
            raise ValueError('Error: frame contains no box data.')
        if frame.box is not None and not self.periodic:
            raise ValueError('Error: frame contains unexpected box data.')
        if self.periodic:
            l, a = v2la(frame.box)
            self.cell_lengths[self.index] = l
            self.cell_angles[self.index] = a
        
    def close(self):
        self.root.close()
        
    def __enter__(self):
        return self
    
    def __exit__(self, *args):
        self.close()

def la2v(lengths, angles):
    alpha = angles[0] * np.pi / 180
    beta = angles[1] * np.pi / 180
    gamma = angles[2] * np.pi / 180

    a = np.array([lengths[0], 0.0, 0.0])
    b = np.array([lengths[1]*np.cos(gamma), lengths[1]*np.sin(gamma), 0.0])
    cx = lengths[2]*np.cos(beta)
    cy = lengths[2]*(np.cos(alpha) - np.cos(beta)*np.cos(gamma)) / np.sin(gamma)
    cz = np.sqrt(lengths[2]*lengths[2] - cx*cx - cy*cy)
    c = np.array([cx,cy,cz])

    v = np.array((a,b,c))
    # Make sure that all vector components that are _almost_ 0 are set exactly
    # to 0
    tol = 1e-6
    v[np.logical_and(v>-tol, v<tol)] = 0.0

    return v

def v2la(vectors):
    a = vectors[0]
    b = vectors[1]
    c = vectors[2]

    a_length = np.sqrt(np.sum(a*a))
    b_length = np.sqrt(np.sum(b*b))
    c_length = np.sqrt(np.sum(c*c))

    alpha = np.arccos(np.dot(b, c) / (b_length * c_length))
    beta = np.arccos(np.dot(c, a) / (c_length * a_length))
    gamma = np.arccos(np.dot(a, b) / (a_length * b_length))

    alpha = alpha * 180.0 / np.pi
    beta = beta * 180.0 / np.pi
    gamma = gamma * 180.0 / np.pi

    return [a_length, b_length, c_length], [alpha, beta, gamma]

def nc_open(filename, mode='r'):
    """
    Open an Amber netcdf trajectory file.
    """
    if not mode in ["r", "w"]:
        raise ValueError('Error: mode must be "r" or "w".')
    if mode == 'r':
        return NCFileReader(filename)
    else:
        return NCFileWriter(filename)
